"""mse_lib_sgx.error module."""


class CryptoError(Exception):
    """Internal cryptographic error."""


class UserError(Exception):
    """User error in the code of the enclave."""


class InternalError(Exception):
    """Internal generic error."""


class SGXError(Exception):
    """Error related to SGX."""


class SecurityError(Exception):
    """Potential security issue."""


class InteruptError(Exception):
    """Interrupt HTTP server."""



"""mse_lib_sgx.import_hook module."""

import os
import sys
from importlib.abc import Loader, MetaPathFinder
from importlib.machinery import ModuleSpec
from importlib.util import spec_from_file_location
from types import ModuleType
from typing import Optional, Sequence, Union, cast

from mse_lib_crypto.xsalsa20_poly1305 import decrypt

from mse_lib_sgx.error import CryptoError


class CipheredMetaFinder(MetaPathFinder):
    """CipheredMetaFinder class."""

    def __init__(self, key: bytes) -> None:
        """Init constructor of CipheredMetaFinder."""
        self.key: bytes = key

    # pylint: disable=unused-argument
    def find_spec(
        self,
        fullname: str,
        path: Optional[Sequence[Union[bytes, str]]],
        target: Optional[ModuleType] = None,
    ) -> Optional[ModuleSpec]:
        """Find the spec for a module."""
        path = cast(Optional[Sequence[str]], path)

        if not path:
            cwd = os.getcwd()
            if cwd not in sys.path:
                sys.path.append(os.getcwd())
            path = sys.path

        if "." in fullname:
            *_, name = fullname.split(".")
        else:
            name = fullname

        for entry in path:
            if os.path.isdir(os.path.join(entry, name)):
                # this module has child modules
                _filename = os.path.join(entry, name, "__init__.py.enc")
                filename = os.path.join(entry, name, "__init__.py")
                submodule_locations = [os.path.join(entry, name)]
            else:
                _filename = os.path.join(entry, name + ".py.enc")
                filename = os.path.join(entry, name + ".py")
                submodule_locations = None

            if os.path.exists(_filename):
                # print("found encrypted module: ", _filename)

                # handle this encrypted file with the Cosmian loader
                return spec_from_file_location(
                    fullname,
                    filename,
                    loader=CipheredLoader(_filename, self.key),
                    submodule_search_locations=submodule_locations,
                )

            if os.path.exists(filename):
                # not us, use the standard loader
                return None

        return None  # we don't know how to import this


class CipheredLoader(Loader):
    """CipheredLoader class."""

    def __init__(self, filename: str, key: bytes) -> None:
        """Init constructor of CipheredLoader."""
        self.filename: str = filename
        self.key: bytes = key

    def create_module(self, spec):
        """Create the module object from the given specification."""
        return None  # use default module creation semantics

    def exec_module(self, module):
        """Initialize the given module object."""
        with open(self.filename, "rb") as f:
            ciphered_module = f.read()
            try:
                plain_module = decrypt(
                    encrypted_data=ciphered_module, key=self.key
                ).decode("utf-8")
            except CryptoError as exc:
                raise CryptoError(
                    f"Failed to decrypt python file: {self.filename}"
                ) from exc
            # pylint: disable=exec-used
            exec(plain_module, vars(module))

    def module_repr(self, module):
        """Return a module's repr.

        Used by the module type when the method does not raise
        NotImplementedError.

        This method is deprecated.

        """
        # The exception will cause ModuleType.__repr__ to ignore this method.
        raise NotImplementedError


def import_set_key(key: bytes) -> None:
    """Configure import hook to decrypt Python modules with `key`."""
    sys.meta_path.insert(0, CipheredMetaFinder(key))



"""mse_lib_sgx.http_server module."""

import json
import logging
import ssl
import threading
import time
from http.server import BaseHTTPRequestHandler, HTTPServer
from typing import Optional
from uuid import UUID

from mse_lib_crypto.seal_box import unseal

from mse_lib_sgx import globs
from mse_lib_sgx.base64url import base64url_decode
from mse_lib_sgx.certificate import Certificate
from mse_lib_sgx.error import CryptoError


class SGXHTTPRequestHandler(BaseHTTPRequestHandler):
    """SGX HTTP server to complete application config with secrets params."""

    def do_GET(self) -> None:
        """GET /."""
        msg: bytes = b"Waiting for code secret key and/or ssl private key..."
        self.send_response(200)
        self.send_header("Content-Length", str(len(msg)))
        # We send an extra header to easily know when the configuration server is up
        self.send_header("Mse-Status", "Waiting")
        self.end_headers()
        self.wfile.write(msg)

    def do_POST(self) -> None:
        """POST /."""
        logging.info("Processing POST query...")
        content_length: int = int(self.headers.get("Content-Length", 0))
        body = self.rfile.read(content_length)

        # body is a json withthese fields:
        # - uuid
        # - (optional) ssl_private_key
        # - (optional) app_secrets
        # - (optional) app_sealed_secrets
        # - code_secret_key
        try:
            data = json.loads(body.decode("utf8"))

            if app_secrets := data.get("app_secrets"):
                globs.SECRETS_PATH.parent.mkdir(parents=True, exist_ok=True)
                globs.SECRETS_PATH.write_bytes(json.dumps(app_secrets).encode("utf-8"))

            if app_sealed_secrets := data.get("app_sealed_secrets"):
                globs.SEALED_SECRETS_PATH.parent.mkdir(parents=True, exist_ok=True)
                globs.SEALED_SECRETS_PATH.write_bytes(
                    json.dumps(
                        json.loads(
                            unseal(
                                base64url_decode(app_sealed_secrets),
                                globs.ENCLAVE_SK_PATH.read_bytes(),
                            )
                        )
                    ).encode("utf-8")
                )

            # Do not process queries which have not the `uuid` data field
            # Probably a robot
            if UUID(data["uuid"]) != globs.ID:
                self.send_response_only(401)
                self.end_headers()
                return

            if globs.NEED_SSL_PRIVATE_KEY:
                globs.SSL_PRIVATE_KEY = data["ssl_private_key"]

            if "code_secret_key" in data:
                globs.CODE_SECRET_KEY = bytes.fromhex(data["code_secret_key"])

                if len(globs.CODE_SECRET_KEY) != 32:
                    raise CryptoError("Incorrect key length!")
        # might be: KeyError, ValueError, json.JSONDecodeError, CryptoError
        except Exception as exc:  # pylint: disable=broad-exception-caught
            logging.error(exc)
            self.send_response_only(401)
            self.end_headers()
            return

        self.send_response_only(200)
        self.end_headers()
        globs.EXIT_EVENT.set()


def serve(
    hostname: str,
    port: int,
    certificate: Certificate,
    app_id: UUID,
    need_ssl_private_key: bool,
    timeout: Optional[int],
):
    """Serve simple SGX HTTP server."""
    globs.NEED_SSL_PRIVATE_KEY = need_ssl_private_key
    globs.ID = app_id

    httpd = HTTPServer((hostname, port), SGXHTTPRequestHandler)

    ctx = ssl.SSLContext(ssl.PROTOCOL_TLS_SERVER)
    ctx.load_cert_chain(
        certfile=str(certificate.cert_path.resolve()),
        keyfile=str(certificate.key_path.resolve()),
    )

    httpd.socket = ctx.wrap_socket(httpd.socket, server_side=True)

    if timeout is not None:
        timer = threading.Timer(interval=timeout, function=kill)
        timer.start()

        threading.Thread(target=kill_event, args=(httpd, timer)).start()
    else:
        threading.Thread(target=kill_event, args=(httpd, None)).start()

    httpd.serve_forever()


def kill_event(httpd: HTTPServer, timer: Optional[threading.Timer]):
    """Kill HTTP server in a thread if `EXIT_EVENT` is set."""
    while True:
        if globs.EXIT_EVENT.is_set():
            logging.info("Stopping the configuration server...")

            if timer:
                timer.cancel()

            httpd.shutdown()
            return

        time.sleep(1)


def kill():
    """Kill HTTP server by setting `EXIT_EVENT`."""
    globs.EXIT_EVENT.set()



"""mse_lib_sgx.copy module."""

import os
import stat
import sys
from shutil import Error, copy, copy2


# pylint: disable=too-many-branches
def _copytree(
    entries,
    src,
    dst,
    symlinks,
    ignore,
    copy_function,
    ignore_dangling_symlinks,
    dirs_exist_ok=False,
):
    """Copy a directory tree and return the destination directory."""
    if ignore is not None:
        ignored_names = ignore(os.fspath(src), [x.name for x in entries])
    else:
        ignored_names = set()

    os.makedirs(dst, exist_ok=dirs_exist_ok)
    errors = []
    use_srcentry = copy_function is copy2 or copy_function is copy

    for srcentry in entries:
        if srcentry.name in ignored_names:
            continue
        srcname = os.path.join(src, srcentry.name)
        dstname = os.path.join(dst, srcentry.name)
        srcobj = srcentry if use_srcentry else srcname
        try:
            is_symlink = srcentry.is_symlink()
            if is_symlink and os.name == "nt":
                # Special check for directory junctions, which appear as
                # symlinks but we want to recurse.
                lstat = srcentry.stat(follow_symlinks=False)
                # pylint: disable=no-member
                if lstat.st_reparse_tag == stat.IO_REPARSE_TAG_MOUNT_POINT:
                    is_symlink = False
            if is_symlink:
                linkto = os.readlink(srcname)
                if symlinks:
                    # We can't just leave it to `copy_function` because legacy
                    # code with a custom `copy_function` may rely on copytree
                    # doing the right thing.
                    os.symlink(linkto, dstname)
                else:
                    # ignore dangling symlink if the flag is on
                    if not os.path.exists(linkto) and ignore_dangling_symlinks:
                        continue
                    # otherwise let the copy occur. copy2 will raise an error
                    if srcentry.is_dir():
                        copytree(
                            srcobj,
                            dstname,
                            symlinks,
                            ignore,
                            copy_function,
                            dirs_exist_ok=dirs_exist_ok,
                        )
                    else:
                        copy_function(srcobj, dstname)
            elif srcentry.is_dir():
                copytree(
                    srcobj,
                    dstname,
                    symlinks,
                    ignore,
                    copy_function,
                    dirs_exist_ok=dirs_exist_ok,
                )
            else:
                # Will raise a SpecialFileError for unsupported file types
                copy_function(srcobj, dstname)
        # catch the Error from the recursive copytree so that we can
        # continue with other files
        except Error as err:
            errors.extend(err.args[0])
        except OSError as why:
            errors.append((srcname, dstname, str(why)))

    if errors:
        raise Error(errors)
    return dst


def copytree(
    src,
    dst,
    symlinks=False,
    ignore=None,
    copy_function=copy2,
    ignore_dangling_symlinks=False,
    dirs_exist_ok=False,
):
    """Recursively copy a directory tree and return the destination directory.

    dirs_exist_ok dictates whether to raise an exception in case dst or any
    missing parent directory already exists.

    If exception(s) occur, an Error is raised with a list of reasons.

    If the optional symlinks flag is true, symbolic links in the
    source tree result in symbolic links in the destination tree; if
    it is false, the contents of the files pointed to by symbolic
    links are copied. If the file pointed by the symlink doesn't
    exist, an exception will be added in the list of errors raised in
    an Error exception at the end of the copy process.

    You can set the optional ignore_dangling_symlinks flag to true if you
    want to silence this exception. Notice that this has no effect on
    platforms that don't support os.symlink.

    The optional ignore argument is a callable. If given, it
    is called with the `src` parameter, which is the directory
    being visited by copytree(), and `names` which is the list of
    `src` contents, as returned by os.listdir():

        callable(src, names) -> ignored_names

    Since copytree() is called recursively, the callable will be
    called once for each directory that is copied. It returns a
    list of names relative to the `src` directory that should
    not be copied.

    The optional copy_function argument is a callable that will be used
    to copy each file. It will be called with the source path and the
    destination path as arguments. By default, copy2() is used, but any
    function that supports the same signature (like copy()) can be used.

    """
    sys.audit("shutil.copytree", src, dst)
    with os.scandir(src) as itr:
        entries = list(itr)
    return _copytree(
        entries=entries,
        src=src,
        dst=dst,
        symlinks=symlinks,
        ignore=ignore,
        copy_function=copy_function,
        ignore_dangling_symlinks=ignore_dangling_symlinks,
        dirs_exist_ok=dirs_exist_ok,
    )



"""mse_lib_sgx.base64url module."""

import base64


def base64url_encode(value: bytes) -> str:
    """Perform URL safe base64 encode on `value` without padding.

    Parameters
    ----------
    value : bytes
        Value to encode in base64.

    Returns
    -------
    str
        Result of URL safe base64 encode on `value`.

    """
    return base64.urlsafe_b64encode(value).rstrip(b"=").decode("ascii")


def base64url_decode(value: str) -> bytes:
    """Perform URL safe b64decode on `value` badly padded.

    Parameters
    ----------
    value : str
        String to b64decode.

    Returns
    -------
    bytes
        Result of b64decode on `value`.

    """
    modulus: int = len(value) % 4

    # add padding depending on modulus
    if modulus:
        # 3 -> '='
        # 2 -> '=='
        value += "=" * (4 - modulus)

    return base64.urlsafe_b64decode(value)
